@@ -3,28 +3,20 @@ import abc
 import errno
 import logging
 import multiprocessing
-import queue
-import sys
-import threading
 import time
 from collections import deque
 from io import BytesIO
 from struct import unpack_from
 
+from .codecs.opus import OpusDecoder
 from .stream_packets import AVPacket, Packet
 from .util import event_emitter, log_bytes, timeit
 
 try:
     import av
-    import ffmpeg
 except ModuleNotFoundError as err:
     print(err)
 
-try:
-    from opuslib import Decoder
-except Exception as err:
-    print(err)
-
 _LOGGER = logging.getLogger(__name__)
 
 
@@ -58,6 +50,7 @@ class AVHandler():
             self._receiver._v_header = v_header
             self._v_stream = AVStream("video", v_header, self._receiver.handle_video)
             self._a_stream = AVStream("audio", a_header, self._receiver.handle_audio)
+            self._receiver.get_audio_config(a_header)
 
     def add_packet(self, packet):
         """Add Packet."""
@@ -145,39 +138,6 @@ class AVStream():
 
         if self._type not in [AVStream.TYPE_VIDEO, AVStream.TYPE_AUDIO]:
             raise ValueError("Invalid Type")
-        if self._type == AVStream.TYPE_AUDIO:
-            self._audio_config = {}
-            self._audio_decoder = None
-            self.get_audio_config()
-
-    def get_audio_config(self):
-        """Get Audio config from header."""
-        if self._type != AVStream.TYPE_AUDIO:
-            raise RuntimeError("Type is not Audio")
-        if sys.modules.get("opuslib") is None:
-            _LOGGER.info("Opuslib not found. Ignoring audio.")
-            return
-        self._audio_config = {
-            "channels": self._header[0],
-            "bits": self._header[1],
-            "rate": unpack_from("!I", self._header, 2)[0],
-            "frame_size": unpack_from("!I", self._header, 6)[0],
-            "unknown": unpack_from("!I", self._header, 10)[0],
-        }
-        _LOGGER.info("Audio Config: %s", self._audio_config)
-        self._audio_decoder = Decoder(self._audio_config['rate'], self._audio_config['channels'])
-
-    def audio_decode(self, packet: AVPacket):
-        if not self._audio_config:
-            return
-
-        assert len(packet.data) % packet.frame_size_audio == 0
-        buf = bytearray()
-        for count in range(0, packet.frame_length_src):
-            start = count * packet.frame_size_audio
-            end = (count + 1) * packet.frame_size_audio
-            buf.extend(self._audio_decoder.decode(packet.data[start:end], self._audio_config['frame_size']))
-        return buf
 
     def handle(self, packet: AVPacket):
         """Handle Packet."""
@@ -188,8 +148,7 @@ class AVStream():
             self._received = 1
 
         if self._type == AVStream.TYPE_AUDIO:
-            data = self.audio_decode(packet)
-            self._callback(data)
+            self._callback(packet)
             return
         # New Video Frame
         if packet.frame_index != self.frame:
@@ -276,6 +235,8 @@ class AVReceiver(abc.ABC):
         #_LOGGER.debug(f"Frame: Key:{frame.key_frame}, Interlaced:{frame.interlaced_frame} Pict:{frame.pict_type}")
         if to_rgb:
             frame = frame.reformat(frame.width, frame.height, "rgb24")
+        elif frame.format.name == "nv12":  # HW Decode will output NV12 frames
+            frame = frame.reformat(format="yuv420p")
         return frame
 
     def find_video_decoder(video_format="h264", use_hw=False):
@@ -316,8 +277,24 @@ class AVReceiver(abc.ABC):
 
     def __init__(self, session):
         self._session = session
-        self.a_cb = None
         self.codec = None
+        self.audio_decoder = None
+        self.audio_config = {}
+
+    def get_audio_config(self, header: bytes):
+        """Get Audio config from header."""
+        self.audio_config = {
+            "channels": header[0],
+            "bits": header[1],
+            "rate": unpack_from("!I", header, 2)[0],
+            "frame_size": unpack_from("!I", header, 6)[0],
+            "unknown": unpack_from("!I", header, 10)[0],
+        }
+        _LOGGER.info("Audio Config: %s", self.audio_config)
+
+        if not self.audio_decoder:
+            self.audio_decoder = OpusDecoder(self.audio_config['rate'], self.audio_config['channels'])
+            event_emitter.emit("audio_config")
 
     def get_video_codec(self):
         self.codec = AVReceiver.video_codec(video_format=self._session.video_format, use_hw=self._session.use_hw)
@@ -334,10 +311,29 @@ class AVReceiver(abc.ABC):
         frame = AVReceiver.video_frame(buf, self.codec)
         return frame
 
+    def decode_audio_frame(self, packet: AVPacket) -> bytes:
+        """Return Audio Frame."""
+        if not self.audio_config:
+            return None
+
+        assert len(packet.data) % packet.frame_size_audio == 0
+        buf = bytearray()
+        for count in range(0, packet.frame_length_src):
+            start = count * packet.frame_size_audio
+            end = (count + 1) * packet.frame_size_audio
+            data = packet.data[start:end]
+            _buf = self.audio_decoder.decode(data)
+            buf.extend(_buf)
+        return buf
+
     def get_video_frame(self):
         """Return Video Frame."""
         raise NotImplementedError
 
+    def get_audio_frame(self):
+        """Return Audio Frame."""
+        raise NotImplementedError
+
     def handle_video(self, buf: bytes):
         """Handle video frame."""
         raise NotImplementedError
@@ -356,10 +352,7 @@ class QueueReceiver(AVReceiver):
     def __init__(self, session):
         super().__init__(session)
         self.v_queue = deque(maxlen=10)
-        self.lock = threading.Lock()
-
-    def add_audio_cb(self, cb):
-        self.a_cb = cb
+        self.a_queue = deque(maxlen=10)
 
     def start(self):
         self.get_video_codec()
@@ -376,108 +369,40 @@ class QueueReceiver(AVReceiver):
         except IndexError:
             return None
 
+    def get_audio_frame(self):
+        try:
+            frame = self.a_queue.popleft()
+            return frame
+        except IndexError:
+            return None
+
     def handle_video(self, buf):
         frame = self.decode_video_frame(buf)
         if frame is None:
             return
-        if self.queue_size >= self.v_queue.maxlen:
-            _LOGGER.warning("AV Receiver max queue size exceeded")
+        if len(self.v_queue) >= self.v_queue.maxlen:
+            _LOGGER.warning("AV Receiver max video queue size exceeded")
             self.v_queue.clear()
         self.v_queue.append(frame)
-        event_emitter.emit('frame')
+        event_emitter.emit('video_frame')
 
-    def handle_audio(self, buf):
-        if self.a_cb is not None:
-            self.a_cb(buf)
-
-    @property
-    def queue_size(self):
-        return len(self.v_queue)
-
-
-# class ProcessReceiver(AVReceiver):
-#     """Uses Multiprocessing. Seems to be slower than threaded."""
-#     class FrameMinimal():
-#         def __init__(self, plane, width, height):
-#             self.planes = [plane]
-#             self.width = width
-#             self.height = height
-
-#     def process(q_in, q_out, video_format="h264", use_hw=False):
-#         codec = AVReceiver.video_codec(video_format=video_format, use_hw=use_hw)
-
-#         _LOGGER.info("Process Started")
-#         while True:
-#             frame = None
-#             try:
-#                 buf = q_in.get_nowait()
-#             except queue.Empty:
-#                 continue
-#             frame = AVReceiver.video_frame(buf, codec)
-#             if frame is None:
-#                 continue
-#             try:
-#                 q_out.put_nowait((bytearray(frame.planes[0]), frame.width, frame.height))
-#                 frame = None
-#                 event_emitter.emit("frame")
-#             except BrokenPipeError:
-#                 break
-
-#     def listener(session, queue):
-#         while not session.is_stopped:
-#             if not queue.empty():
-#                 event_emitter.emit("frame")
-#                 break
-#             else:
-#                 time.sleep(0.001)
-
-#     def __init__(self, session):
-#         super().__init__(session)
-#         self.pipe1, self.pipe2 = multiprocessing.Pipe()
-#         self.manager = multiprocessing.Manager()
-#         self.v_queue_in = self.manager.Queue(100)
-#         self.v_queue = self.manager.Queue(10)
-#         self.lock = self.manager.Lock()
-#         self._worker = None
-
-#     def start(self):
-#         self._worker = multiprocessing.Process(
-#             target=ProcessReceiver.process,
-#             args=(self.v_queue_in, self.v_queue, self._session.video_format, self._session.use_hw),
-#             daemon=True,
-#         )
-#         self._worker.start()
-#         _LOGGER.info("Process Start")
-#         self._listener = threading.Thread(target=ProcessReceiver.listener, args=(self._session, self.v_queue))
-#         self._listener.start()
-#         self.notify_started()
-
-#     def get_video_frame(self):
-#         frame = None
-#         try:  
-#             frame = self.v_queue.get_nowait()
-#             frame = ProcessReceiver.FrameMinimal(frame[0], frame[1], frame[2])
-#         except queue.Empty:
-#             pass
-#         return frame
-
-#     def handle_video(self, buf):
-#         self.v_queue_in.put_nowait(buf)
-
-#     def handle_audio(self, buf):
-#         if self.a_cb is not None:
-#             self.a_cb(buf)
-
-#     def close(self):
-#         if self._worker:
-#             self._worker.terminate()
-#             self._worker.join()
-#             self._worker.close()
+    def handle_audio(self, packet):
+        frame = self.decode_audio_frame(packet)
+        if frame is None:
+            return
+        if len(self.a_queue) >= self.a_queue.maxlen:
+            _LOGGER.warning("AV Receiver max audio queue size exceeded")
+            self.a_queue.clear()
+        self.a_queue.append(frame)
+        event_emitter.emit('audio_frame')
 
 
 class AVFileReceiver(AVReceiver):
     """Writes AV to file."""
+
     def process(pipe, av_file):
+        import ffmpeg
+
         video = ffmpeg.input('pipe:', format='h264')
         audio = ffmpeg.input('pipe:', format='s16le', ac=2, ar=48000)
         #joined = ffmpeg.concat(video, audio, v=1, a=1).node
