@@ -1,10 +1,5 @@
 """Device Grid Widget."""
-
-import requests
-from pyps4_2ndscreen.ddp import get_status, search
-from pyps4_2ndscreen.media_art import (BASE_IMAGE_URL, BASE_URL,
-                                       DEFAULT_HEADERS, ResultItem,
-                                       get_region_codes)
+from pyremoteplay.ddp import DEFAULT_STANDBY_DELAY, get_status, search
 from PySide6 import QtCore, QtGui, QtWidgets
 from PySide6.QtCore import Qt
 
@@ -14,14 +9,15 @@ class DeviceButton(QtWidgets.QPushButton):
     COLOR_LIGHT = "#FFFFFF"
     COLOR_BG = "#E9ECEF"
 
-    def __init__(self, main_window, host):
+    def __init__(self, main_window, device):
         super().__init__()
         self.info = ""
         self.main_window = main_window
-        self.host = host
+        self.device = device
+        self.status = device.status
         self.info_show = False
         self.menu = QtWidgets.QMenu(self)
-        self.clicked.connect(lambda: self.main_window.connect_host(self.host))
+        self.clicked.connect(lambda: self.main_window.connect_host(self.status))
         self.clicked.connect(lambda: self.setEnabled(False))
         self.text_color = self.COLOR_DARK
         self.bg_color = self.COLOR_BG
@@ -42,7 +38,7 @@ class DeviceButton(QtWidgets.QPushButton):
         self.action_power.triggered.connect(self.toggle_power)
 
     def set_style(self):
-        if self.host["status_code"] == 200:
+        if self.status["status_code"] == 200:
             self.border_color = ("#6EA8FE", "#0D6EFD")
         else:
             self.border_color = ("#FEB272", "#FFC107")
@@ -61,38 +57,26 @@ class DeviceButton(QtWidgets.QPushButton):
         ))
 
     def update_state(self, state):
-        cur_id = self.host.get("running-app-titleid")
+        cur_id = self.status.get("running-app-titleid")
         new_id = state.get("running-app-titleid")
-        self.host = state
+        cur_status = self.status.get('status_code')
+        new_status = state.get('status_code')
+        self.status = state
         self.get_info()
         self.get_text()
         if cur_id != new_id:
             self.set_image()
+        if cur_status == 200 and new_status != cur_status:
+            # Disable Wakeup since device won't do anything right away.
+            self.action_power.setDisabled(True)
+            QtCore.QTimer.singleShot(DEFAULT_STANDBY_DELAY * 1000, self.enable_toggle_power)
         self.set_style()
 
     def set_image(self):
-        def get_image(title_id):
-            image = None
-            codes = get_region_codes("United States")
-            data_url = BASE_URL.format(codes[1], codes[0], title_id)
-            image_url = BASE_IMAGE_URL.format(data_url)
-            resp = requests.get(data_url, headers=DEFAULT_HEADERS)
-            if not resp:
-                return None
-            data = resp.json()
-            if data.get("gameContentTypesList") is None or data.get("title_name") is None:
-                return None
-            item = ResultItem(title_id, image_url, data)
-            if item is None or item.cover_art is None:
-                return None
-            resp = requests.get(item.cover_art)
-            image = resp.content
-            return image
-
         self.bg_color = self.COLOR_BG
-        title_id = self.host.get("running-app-titleid")
+        title_id = self.status.get("running-app-titleid")
         if title_id:
-            image = get_image(title_id)
+            image = self.device.image
             if image is not None:
                 pix = QtGui.QPixmap()
                 pix.loadFromData(image)
@@ -107,9 +91,10 @@ class DeviceButton(QtWidgets.QPushButton):
                     self.text_color = self.COLOR_DARK
         else:
             self.setIcon(QtGui.QIcon())
+            self.text_color = self.COLOR_DARK
 
     def calc_contrast(self, hex_color):
-        colors = (self.text_color, hex_color)
+        colors = (self.COLOR_DARK, hex_color)
         lum = []
         for color in colors:
             lum.append(self.calc_luminance(color))
@@ -137,16 +122,16 @@ class DeviceButton(QtWidgets.QPushButton):
         return luminance
 
     def get_text(self):
-        device_type = self.host['host-type']
-        if self.host['host-type'] == "PS4":
+        device_type = self.status['host-type']
+        if self.status['host-type'] == "PS4":
             device_type = "PlayStation 4"
-        elif self.host['host-type'] == "PS5":
+        elif self.status['host-type'] == "PS5":
             device_type = "PlayStation 5"
-        app = self.host.get('running-app-name')
+        app = self.status.get('running-app-name')
         if not app:
-            app = "Idle" if self.host["status_code"] == 200 else "Standby" 
+            app = "Idle" if self.status["status_code"] == 200 else "Standby" 
         self.main_text = (
-            f"{self.host['host-name']}\n"
+            f"{self.status['host-name']}\n"
             f"{device_type}\n\n"
             f"{app}"
         )
@@ -155,18 +140,18 @@ class DeviceButton(QtWidgets.QPushButton):
 
     def get_info(self):
         self.info = (
-            f"Type: {self.host['host-type']}\n"
-            f"Name: {self.host['host-name']}\n"
-            f"IP Address: {self.host['host-ip']}\n"
-            f"Mac Address: {self.host['host-id']}\n\n"
-            f"Status: {self.host['status']}\n"
-            f"Playing: {self.host.get('running-app-name')}"
+            f"Type: {self.status['host-type']}\n"
+            f"Name: {self.status['host-name']}\n"
+            f"IP Address: {self.status['host-ip']}\n"
+            f"Mac Address: {self.status['host-id']}\n\n"
+            f"Status: {self.status['status']}\n"
+            f"Playing: {self.status.get('running-app-name')}"
         )
 
     def contextMenuEvent(self, event):
         text = "View Info" if not self.info_show else "Hide Info"
         self.action_info.setText(text)
-        if self.host['status_code'] == 200:
+        if self.status['status_code'] == 200:
             self.action_power.setText("Standby")
         else:
             self.action_power.setText("Wakeup")
@@ -178,118 +163,74 @@ class DeviceButton(QtWidgets.QPushButton):
         self.info_show = not self.info_show
 
     def toggle_power(self):
-        if self.host['status_code'] == 200:
-            self.main_window.standby_host(self.host)
+        if self.status['status_code'] == 200:
+            self.main_window.standby_host(self.status)
         else:
-            self.main_window.wakeup_host(self.host)
+            self.main_window.wakeup_host(self.status)
 
-
-class DeviceSearch(QtCore.QObject):
-    finished = QtCore.Signal()
-
-    def __init__(self, main_window):
-        super().__init__()
-        self.main_window = main_window
-        self.hosts = []
-
-    def get_hosts(self):
-        manual_hosts = self.main_window.options.options.get("devices")
-        self.hosts = search()
-        if manual_hosts:
-            found = []
-            if self.hosts:
-                for item in self.hosts:
-                    found.append(item.get("host-ip"))
-            for _host in manual_hosts:
-                if _host in found:
-                    continue
-                host = get_status(_host)
-                if host:
-                    self.hosts.append(host)
-        self.finished.emit()
+    def enable_toggle_power(self):
+        self.action_power.setDisabled(False)
 
 
 class DeviceGridWidget(QtWidgets.QWidget):
 
-    def __init__(self, main_window):
-        super().__init__(main_window)
+    def __init__(self, parent, main_window):
+        super().__init__(parent)
         self.main_window = main_window
         self.layout = QtWidgets.QGridLayout(self)
         self.layout.setColumnMinimumWidth(0, 100)
-        self.widgets = []
+        self.widgets = {}
         self.setStyleSheet("QPushButton {padding: 50px 25px;}")
-        self.searcher = DeviceSearch(self.main_window)
-        self.timer = QtCore.QTimer()
-        self.timer.timeout.connect(self.discover)
-        self.wait_timer = None
-        self.thread = QtCore.QThread(self)
-        self.searcher.moveToThread(self.thread)
-        self.thread.started.connect(self.searcher.get_hosts)
-        self.searcher.finished.connect(lambda: self.create_grid(self.searcher.hosts))
-        self.searcher.finished.connect(self.thread.quit)
-        self._is_startup = True
 
     def add(self, button, row, col):
         self.layout.setRowStretch(row, 6)
         self.layout.addWidget(button, row, col, Qt.AlignCenter)
-        self.widgets.append(button)
+        self.widgets[button.device.host] = button
 
-    def create_grid(self, hosts):
+    def create_grid(self, devices):
+        all_devices = []
         if self.widgets:
-            ip_addresses = {}
-            if hosts:
-                for item in hosts:
-                    ip_address = item["host-ip"]
-                    ip_addresses[ip_address] = item
-                    hosts.remove(item)
-            for widget in self.widgets:
+            for widget in self.widgets.values():
                 self.layout.removeWidget(widget)
-                new_state = ip_addresses.get(widget.host["host-ip"])
-                if not new_state:
-                    widget.setParent(None)
-                    widget.deleteLater()
-                else:
-                    widget.update_state(new_state)
-                    hosts.append(widget)
-            self.widgets = []
+        widget = None
+        for ip_address in devices.keys():
+            if ip_address not in self.widgets:
+                all_devices.append(devices[ip_address])
+            else:
+                widget = self.widgets.pop(ip_address)
+                all_devices.append(widget)
+                widget.update_state(widget.device.status)
+        for widget in self.widgets.values():
+            widget.setParent(None)
+            widget.deleteLater()
+        self.widgets = {}
+
         max_cols = 3
-        if hosts:
-            for index, host in enumerate(hosts):
+        if all_devices:
+            for index, device in enumerate(all_devices):
                 col = index % max_cols
                 row = index // max_cols
-                if isinstance(host, dict):
-                    button = DeviceButton(self.main_window, host)
+                if isinstance(device, DeviceButton):
+                    button = device
                 else:
-                    button = host
+                    button = DeviceButton(self.main_window, device)
                 self.add(button, row, col)
             if not self.main_window.toolbar.options.isChecked() \
                     and not self.main_window.toolbar.controls.isChecked():
                 self.show()
             self.main_window.center_text.hide()
-        else:
-            if self._is_startup:
-                self.main_window.startup_check_grid()
-                self._is_startup = False
-            else:
-                self.main_window.center_text.hide()
-
-    def discover(self):
-        self.thread.start()
-
-    def start_timer(self):
-        self.timer.start(5000)
 
     def session_stop(self):
         if self.main_window.toolbar.refresh.isChecked():
-            self.start_timer()
-        self.wait_timer = QtCore.QTimer.singleShot(10000, self.enable_buttons)
+            self.start_update()
+        QtCore.QTimer.singleShot(10000, self.enable_buttons)
 
     def enable_buttons(self):
-        for button in self.widgets:
+        for button in self.widgets.values():
             button.setDisabled(False)
 
+    def start_update(self):
+        self.main_window.async_handler.poll()
+
     def stop_update(self):
-        self.timer.stop()
-        if self.wait_timer:
-            self.wait_timer.stop()
-        self.thread.quit()
+        self.main_window.async_handler.stop_poll()
