@@ -1,4 +1,6 @@
-from pyps4_2ndscreen.ddp import get_status
+"""Options Window."""
+import socket
+
 from pyremoteplay.av import AVReceiver
 from pyremoteplay.const import RESOLUTION_PRESETS, Quality
 from pyremoteplay.oauth import LOGIN_URL, get_user_account
@@ -8,8 +10,10 @@ from pyremoteplay.util import (add_profile, add_regist_data, get_mapping,
                                write_options, write_profiles)
 from PySide6 import QtCore, QtWidgets
 from PySide6.QtCore import Qt
+from PySide6.QtMultimedia import QMediaDevices
 
 from .util import label, message, spacer
+from .widgets import AnimatedToggle
 
 
 class ControlsTable(QtWidgets.QTableWidget):
@@ -85,7 +89,7 @@ class ControlsWidget(QtWidgets.QWidget):
         super().__init__(main_window)
         self.main_window = main_window
         self.mapping = None
-        self.options = None
+        self.options = {}
         self.selected_map = ""
         self.layout = QtWidgets.QGridLayout(self, alignment=Qt.AlignTop)
         self.layout.setColumnMinimumWidth(0, 30)
@@ -102,20 +106,20 @@ class ControlsWidget(QtWidgets.QWidget):
         header = self.table.horizontalHeader()       
         header.setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
         header.setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeToContents)
-        self.left_joystick = QtWidgets.QCheckBox("Show Left Joystick", self)
-        self.right_joystick = QtWidgets.QCheckBox("Show Right Joystick", self)
+        self.left_joystick = AnimatedToggle("Show Left Joystick", self)
+        self.right_joystick = AnimatedToggle("Show Right Joystick", self)
         self.reset = QtWidgets.QPushButton("Reset to Default")
         self.clear = QtWidgets.QPushButton("Clear")
         self.cancel = QtWidgets.QPushButton("Cancel")
         self.init_controls()
         self.instructions()
-        self.layout.addWidget(self.left_joystick, 0, 0)
-        self.layout.addWidget(self.right_joystick, 0, 1)
-        self.layout.addWidget(self.reset, 1, 0)
-        self.layout.addWidget(self.clear, 1, 1)
-        self.layout.addWidget(self.cancel, 1, 2)
-        self.layout.addWidget(self.table, 2, 0, 1, 3)
-        self.layout.addWidget(self.label, 2, 3)
+        self.layout.addWidget(self.left_joystick, 0, 0, 1, 2)
+        self.layout.addWidget(self.right_joystick, 1, 0, 1, 2)
+        self.layout.addWidget(self.reset, 2, 0)
+        self.layout.addWidget(self.clear, 2, 1)
+        self.layout.addWidget(self.cancel, 2, 2)
+        self.layout.addWidget(self.table, 3, 0, 1, 3)
+        self.layout.addWidget(self.label, 3, 3)
         self.cancel.hide()
         self.clear.hide()
         self.left_joystick.clicked.connect(lambda: self.click_joystick("left"))
@@ -306,6 +310,11 @@ class OptionsWidget(QtWidgets.QWidget):
 
         # self.layout.setRowMinimumHeight(3, 20)
         # self.layout.setRowStretch(4, 1)
+        self.audio_output = None
+        self.audio_devices = {}
+        self.get_audio_devices()
+        self._media_devices = QMediaDevices()
+        self._media_devices.audioOutputsChanged.connect(self.get_audio_devices)
 
         self.init_options()
         self.load_options()
@@ -315,6 +324,7 @@ class OptionsWidget(QtWidgets.QWidget):
                 raise RuntimeError("Failed to set options")
         self.set_profiles()
         self.set_devices()
+        self.main_window.add_devices(self.options["devices"])
 
     def init_options(self):
         set_account = QtWidgets.QPushButton("Select Account")
@@ -339,17 +349,21 @@ class OptionsWidget(QtWidgets.QWidget):
             "try adding the IP Address below.",
             wrap=True,
         )
+
         self.quality = QtWidgets.QComboBox(self)
         self.quality.addItems([item.name for item in Quality])
         self.quality.currentTextChanged.connect(self.change_quality)
+        self.use_opengl = AnimatedToggle("Use OpenGL", self)
+        self.use_opengl.stateChanged.connect(self.change_opengl)
+        self.use_opengl.setToolTip("Recommended if using Hardware Decoding")
         self.fps = QtWidgets.QComboBox(self)
         self.fps.addItems(["30", "60"])
         self.fps.currentTextChanged.connect(self.change_fps)
-        self.fps_show = QtWidgets.QCheckBox("Show FPS", self)
+        self.fps_show = AnimatedToggle("Show FPS", self)
         self.fps_show.stateChanged.connect(self.change_fps_show)
-        self.fullscreen = QtWidgets.QCheckBox("Show Fullscreen", self)
+        self.fullscreen = AnimatedToggle("Show Fullscreen", self)
         self.fullscreen.stateChanged.connect(self.change_fullscreen)
-        self.use_hw = QtWidgets.QCheckBox("Use Hardware Decoding", self)
+        self.use_hw = AnimatedToggle("Use Hardware Decoding", self)
         self.use_hw.stateChanged.connect(self.change_use_hw)
         self.resolution = QtWidgets.QComboBox(self)
         self.resolution.addItems(list(RESOLUTION_PRESETS.keys()))
@@ -358,8 +372,11 @@ class OptionsWidget(QtWidgets.QWidget):
         self.accounts.itemDoubleClicked.connect(self.change_profile)
         self.devices = QtWidgets.QTreeWidget()
         self.devices.itemClicked.connect(del_device.show)
+        self.audio_output = QtWidgets.QComboBox(self)
+        self.audio_output.addItems(list(self.audio_devices.keys()))
 
         self.add(self.quality, 0, 1, label=label(self, "Quality:"))
+        self.add(self.use_opengl, 0, 2)
         self.add(self.fps, 1, 1, label=label(self, "FPS:"))
         self.add(self.fps_show, 1, 2)
         self.add(self.resolution, 2, 1, label=label(self, "Resolution:"))
@@ -368,15 +385,16 @@ class OptionsWidget(QtWidgets.QWidget):
         self.add(hw_label, 4, 1, label=label(self, "Available Decoder:"))
         self.layout.addWidget(self.use_hw, 4, 2)
         self.layout.addWidget(devices_label, 4, 4, 1, 2)
-        self.layout.addItem(spacer(), 5, 0)
-        self.add(set_account, 6, 0)
-        self.add(add_account, 6, 1)
-        self.add(del_account, 6, 2)
-        self.add(add_device, 6, 4)
-        self.add(del_device, 6, 5)
-        self.layout.addWidget(self.accounts, 7, 0, 3, 3)
-        self.layout.addItem(spacer(), 7, 3)
-        self.layout.addWidget(self.devices, 7, 4, 3, 2)
+        self.add(self.audio_output, 5, 1, width=3, label=label(self, "Audio Output"))
+        self.layout.addItem(spacer(), 6, 0)
+        self.add(set_account, 7, 0)
+        self.add(add_account, 7, 1)
+        self.add(del_account, 7, 2)
+        self.add(add_device, 7, 4)
+        self.add(del_device, 7, 5)
+        self.layout.addWidget(self.accounts, 8, 0, 3, 3)
+        self.layout.addItem(spacer(), 8, 3)
+        self.layout.addWidget(self.devices, 8, 4, 3, 2)
         del_device.hide()
 
     def get_decoder(self):
@@ -386,9 +404,27 @@ class OptionsWidget(QtWidgets.QWidget):
             decoder = "CPU"
         return decoder
 
+    def get_audio_devices(self):
+        devices = QMediaDevices.audioOutputs()
+        default = QMediaDevices.defaultAudioOutput()
+        audio_devices = {f"{default.description()} (Default)": default}
+        for device in devices:
+            if device == default:
+                continue
+            audio_devices[device.description()] = device
+        self.audio_devices = audio_devices
+        if self.audio_output:
+            self.audio_output.clear()
+            self.audio_output.addItems(self.audio_devices)
+        return audio_devices
+
+    def get_audio_device(self):
+        return self.audio_devices.get(self.audio_output.currentText())
+
     def set_options(self) -> bool:
         try:
             self.quality.setCurrentText(str(self.options["quality"]))
+            self.use_opengl.setChecked(self.options["use_opengl"])
             self.fps.setCurrentText(str(self.options["fps"]))
             self.fps_show.setChecked(self.options["show_fps"])
             self.use_hw.setChecked(self.options["use_hw"])
@@ -403,6 +439,7 @@ class OptionsWidget(QtWidgets.QWidget):
     def default_options(self) -> dict:
         options = {
             "quality": "Default",
+            "use_opengl": False,
             "fps": 30,
             "show_fps": False,
             "resolution": "720p",
@@ -421,8 +458,6 @@ class OptionsWidget(QtWidgets.QWidget):
         self.options = options
 
     def set_devices(self):
-        if not self.options.get("devices"):
-            self.options["devices"] = []
         self.devices.clear()
         self.devices.setHeaderLabels(["Devices"])
         for host in self.options["devices"]:
@@ -456,15 +491,20 @@ class OptionsWidget(QtWidgets.QWidget):
             selected[0].setSelected(True)
             selected[0].setText(1, "Yes")
 
-    def add(self, item, row, col, label=None):
-        self.layout.addWidget(item, row, col, Qt.AlignLeft)
+    def add(self, item, row, col, width=2, label=None):
         if label is not None:
+            width -= 1
             self.layout.addWidget(label, row, col - 1, Qt.AlignLeft)
+        self.layout.addWidget(item, row, col, 1, width, Qt.AlignLeft)
 
     def change_quality(self, text):
         self.options["quality"] = text
         write_options(self.options)
 
+    def change_opengl(self):
+        self.options["use_opengl"] = self.use_opengl.isChecked()
+        write_options(self.options)
+
     def change_fps(self, text):
         self.options["fps"] = int(text)
         write_options(self.options)
@@ -504,18 +544,24 @@ class OptionsWidget(QtWidgets.QWidget):
         host = dialog.textValue()
         if not host:
             return
+        try:
+            socket.getaddrinfo(host, None)
+        except socket.gaierror:
+            QtWidgets.QMessageBox.critical(self, "Invalid IP Address", f"Could not find device at: {host}")
+            return
         if host in self.options["devices"]:
             text = "Device is already added."
             message(self.main_window, "Device Already Added", text, "warning")
             return
-        status = get_status(host)
-        if not status:
-            text = f"Could not find device at: {host}."
-            message(self.main_window, "Device not found", text, "warning")
-            return
+        # status = get_status(host)
+        # if not status:
+        #     text = f"Could not find device at: {host}."
+        #     message(self.main_window, "Device not found", text, "warning")
+        #     return
         self.options["devices"].append(host)
         write_options(self.options)
         self.set_devices()
+        self.main_window.add_devices(self.options["devices"])
 
     def delete_device(self):
         items = self.devices.selectedItems()
@@ -526,6 +572,7 @@ class OptionsWidget(QtWidgets.QWidget):
         self.options["devices"].remove(host)
         write_options(self.options)
         self.set_devices()
+        self.main_window.remove_device(host)
 
     def delete_profile(self):
         item = self.accounts.selectedItems()[0]
