@@ -1,24 +1,53 @@
 """Stream Window for GUI."""
-import asyncio
-import sys
-import threading
 import time
 from collections import deque
 from enum import Enum
 
-from pyremoteplay.av import QueueReceiver
+from pyremoteplay.av import AVReceiver
 from pyremoteplay.session import SessionAsync
 from pyremoteplay.util import event_emitter, timeit
-from PySide6 import QtCore, QtGui, QtWidgets
+from PySide6 import QtCore, QtGui, QtMultimedia, QtWidgets
 from PySide6.QtCore import Qt
 
 from .options import ControlsWidget
 from .util import label, message
+from .video import VideoWidget, YUVGLWidget
+
+
+class QtReceiver(AVReceiver):
+
+    def __init__(self, session):
+        super().__init__(session)
+        self.video_signal = None
+        self.audio_signal = None
+        self.rgb = True
+
+    def decode_video_frame(self, buf: bytes) -> bytes:
+        """Decode Video Frame."""
+        frame = AVReceiver.video_frame(buf, self.codec, to_rgb=self.rgb)
+        return frame
+
+    def handle_video(self, buf):
+        frame = self.decode_video_frame(buf)
+        if frame is None:
+            return
+        self.video_signal.emit(frame)
+
+    def handle_audio(self, packet):
+        frame = self.decode_audio_frame(packet)
+        if frame is None:
+            return
+        self.audio_signal.emit(frame)
+
+    def set_signals(self, video_signal, audio_signal):
+        self.video_signal = video_signal
+        self.audio_signal = audio_signal
 
 
 class RPWorker(QtCore.QObject):
     finished = QtCore.Signal()
     started = QtCore.Signal()
+    standby_done = QtCore.Signal()
 
     def __init__(self, main_window):
         super().__init__()
@@ -27,61 +56,64 @@ class RPWorker(QtCore.QObject):
         self.session = None
         self.thread = None
         self.error = ""
+        self.standby_done.connect(self.standby_finished)
         event_emitter.on("stop", self.stop)
 
-    def run(self):
+    def run(self, standby=False):
         if not self.session:
             print("No Session")
             self.stop()
             return
-        if not self.window:
+        if not self.window and not standby:
             print("No Stream Window")
             self.stop()
             return
-        self.thread = threading.Thread(
-            target=self.worker,
-        )
-        self.thread.start()
+        self.session.loop = self.main_window.async_handler.loop
+        self.session.loop.create_task(self.start(standby))
 
-    def stop(self):
+    def stop(self, standby=False):
         if self.session:
             print(f"Stopping Session @ {self.session.host}")
             self.session.stop()
             self.error = self.session.error
+        if standby:
+            self.standby_done.emit()
         self.session = None
         self.window = None
         self.finished.emit()
 
-    def setup(self, window, host, profile, resolution, fps, use_hw, quality):
+    def setup(self, window, host, profile, resolution='720p', fps=60, use_hw=False, quality='default'):
         self.window = window
-        self.session = SessionAsync(host, profile, resolution=resolution, fps=fps, av_receiver=QueueReceiver, use_hw=use_hw, quality=quality)
-        # self.session.av_receiver.add_audio_cb(self.handle_audio)
+        self.session = SessionAsync(host, profile, resolution=resolution, fps=fps, av_receiver=QtReceiver, use_hw=use_hw, quality=quality)
 
-    def worker(self):
-        if sys.platform == "win32":
-            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
-        self.session.loop = asyncio.new_event_loop()
-        task = self.session.loop.create_task(self.start())
+    async def start(self, standby=False):
         print("Session Start")
-        self.session.loop.run_until_complete(task)
-        if self.session:
-            self.session.loop.run_forever()
-        print("Session Finished")
-        task.cancel()
-        self.stop()
-
-    async def start(self):
+        self.session.av_receiver.rgb = False if self.window.use_opengl else True
         status = await self.session.start()
         if not status:
             print("Session Failed to Start")
-            message(None, "Error", self.session.error)
+            # message(None, "Error", self.session.error)
             self.stop()
+            return
         else:
+            self.session.av_receiver.set_signals(self.window.video_frame, self.window.audio_frame)
+            self.window.audio_frame.connect(self.window.next_audio_frame)
+            self.window.video_frame.connect(self.window.video_output.next_video_frame)
             self.started.emit()
+        if standby:
+            await self.session.stream_ready.wait()
+            self.send_standby()
+        elif self.session._stop_event:
+            await self.session._stop_event.wait()
+            print("Session Finished")
 
     def send_standby(self):
         self.session.standby()
-        self.stop()
+        self.stop(standby=True)
+
+    def standby_finished(self):
+        host = self.session.host if self.session else "Unknown"
+        self.main_window.standby_callback(host)
 
     def stick_state(self, stick: str, direction: str = None, value: float = None, point=None):
         if point is not None:
@@ -100,48 +132,27 @@ class RPWorker(QtCore.QObject):
         self.session.controller.button(button, action)
 
 
-class AVProcessor(QtCore.QObject):
-    started = QtCore.Signal()
-    frame = QtCore.Signal()
-    slow = QtCore.Signal()
-
-    def __init__(self, window):
-        super().__init__()
-        self.window = window
-        self.pixmaps = deque()
-        self._set_slow = False
-        event_emitter.on("frame", self.next_frame)
-
-    def next_frame(self):
-        frame = self.window.rp_worker.session.av_receiver.get_video_frame()
-        if frame is None:
-            return
-        image = QtGui.QImage(
-            bytearray(frame.planes[0]),
-            frame.width,
-            frame.height,
-            frame.width * 3,
-            QtGui.QImage.Format_RGB888,
-        )
-        if len(self.pixmaps) > 2:
-            self.pixmaps.clear()
-        self.pixmaps.append(QtGui.QPixmap.fromImage(image))
-
-
 class JoystickWidget(QtWidgets.QFrame):
     def __init__(self, window, left=False, right=False):
         super().__init__(window)
         self.window = window
+        self.grab_outside = False
         self.left = Joystick(self, "left") if left else None
         self.right = Joystick(self, "right") if right else None
         self.layout = QtWidgets.QHBoxLayout(self)
         self.layout.setAlignment(Qt.AlignCenter)
         self.layout.setContentsMargins(0, 0, 0, 0)
         self.setStyleSheet("background-color: rgba(255, 255, 255, 0.4); border-radius:25%;")
+        self.set_cursor()
         for joystick in [self.left, self.right]:
             self.layout.addWidget(joystick)
             joystick.show()
 
+    def set_cursor(self, shape=Qt.SizeAllCursor):
+        cursor = QtGui.QCursor()
+        cursor.setShape(shape)
+        self.setCursor(cursor)
+
     def hide_sticks(self):
         self.left.hide()
         self.right.hide()
@@ -177,6 +188,8 @@ class JoystickWidget(QtWidgets.QFrame):
         self.grab_outside = False
 
     def mouseMoveEvent(self, event):
+        if event.buttons() == QtCore.Qt.NoButton:
+            return
         if self.grab_outside:
             cur_pos = self.mapToGlobal(self.pos())
             global_pos = event.globalPos()
@@ -214,11 +227,12 @@ class Joystick(QtWidgets.QLabel):
         self.grabbed = False
         self.__maxDistance = 50
         self.setStyleSheet("background-color: rgba(0, 0, 0, 0.0)")
-        self.set_default_cursor()
+        self.set_cursor()
+        self.setMouseTracking(True)
 
-    def set_default_cursor(self):
+    def set_cursor(self, shape=Qt.SizeAllCursor):
         cursor = QtGui.QCursor()
-        cursor.setShape(Qt.SizeAllCursor)
+        cursor.setShape(shape)
         self.setCursor(cursor)
 
     def paintEvent(self, event):
@@ -258,32 +272,46 @@ class Joystick(QtWidgets.QLabel):
         if is_center:
             self.grabbed = True
             self.movingOffset = self._boundJoystick(event.pos())
+            point = self.joystickDirection()
+            self.parent.window.rp_worker.stick_state(self.stick, point=point)
             self.update()
         if not self.grabbed:
             self.parent.mousePressEvent(event)
-        return super().mousePressEvent(event)
+        else:
+            self.set_cursor(Qt.ClosedHandCursor)
 
     def mouseReleaseEvent(self, event):
-        if not self.grabbed:
-            self.parent.mouseMoveEvent(event)
-        self.grabbed = False
-        self.movingOffset = QtCore.QPointF(0, 0)
-        self.update()
-        point = self.joystickDirection()
-        self.parent.window.rp_worker.stick_state(self.stick, point=point)
+        if self.grabbed:
+            self.grabbed = False
+            self.movingOffset = QtCore.QPointF(0, 0)
+            point = self.joystickDirection()
+            self.parent.window.rp_worker.stick_state(self.stick, point=point)
+            self.set_cursor(Qt.OpenHandCursor)
+            self.update()
+        else:
+            self.parent.mouseReleaseEvent(event)
 
     def mouseMoveEvent(self, event):
         if self.grabbed:
+            self.set_cursor(Qt.ClosedHandCursor)
             self.movingOffset = self._boundJoystick(event.pos())
-            self.update()
             point = self.joystickDirection()
             self.parent.window.rp_worker.stick_state(self.stick, point=point)
+            self.update()
         else:
+            is_center = self._centerEllipse().contains(event.pos())
+            if is_center:
+                self.set_cursor(Qt.OpenHandCursor)
+            else:
+                self.set_cursor()
             self.parent.mouseMoveEvent(event)
 
 
 class StreamWindow(QtWidgets.QWidget):
     started = QtCore.Signal()
+    fps_update = QtCore.Signal()
+    video_frame = QtCore.Signal(object)
+    audio_frame = QtCore.Signal(object)
 
     def __init__(self, main_window):
         super().__init__()
@@ -293,94 +321,91 @@ class StreamWindow(QtWidgets.QWidget):
         self.setMaximumWidth(self.main_window.screen.virtualSize().width())
         self.setMaximumHeight(self.main_window.screen.virtualSize().height())
         self.setStyleSheet("background-color: black")
-        self.video_output = QtWidgets.QLabel(self, alignment=Qt.AlignCenter)
+        self.video_output = None
         self.audio_output = None
+        self.audio_device = None
+        self.opengl = False
+        self.center_text = QtWidgets.QLabel("Starting Stream...", alignment=Qt.AlignCenter)
+        self.center_text.setWordWrap(True)
+        self.center_text.setStyleSheet("QLabel {color: white;font-size: 24px;}")
         self.layout = QtWidgets.QVBoxLayout(self)
         self.layout.setContentsMargins(0, 0, 0, 0)
-        self.layout.addWidget(self.video_output)
+        self.layout.addWidget(self.center_text)
         self.joystick = JoystickWidget(self, left=True, right=True)
         self.joystick.hide()
         self.input_options = None
         self.fps_label = label(self, "FPS: ")
-        self.av_worker = AVProcessor(self)
-        self.timer = QtCore.QTimer()
-        self.timer.setTimerType(Qt.PreciseTimer)
-        self.timer.timeout.connect(self.new_frame)
-        self.ms_refresh = 0
-        self._video_transform_mode = Qt.SmoothTransformation
+        self.fps_update.connect(self.set_fps)
 
         self.rp_worker = self.main_window.rp_worker
-        self.av_thread = QtCore.QThread()
         self.rp_worker.started.connect(self.show_video)
         self.rp_worker.finished.connect(self.close)
-        self.av_worker.slow.connect(self.av_slow)
-        self.av_worker.moveToThread(self.av_thread)
-        self.av_thread.started.connect(self.start_timer)
-        # self.av_thread.started.connect(self.av_worker.next_frame)
+        event_emitter.on("audio_config", self.init_audio)
 
-    def start(self, host, name, profile, resolution='720p', fps=60, show_fps=False, fullscreen=False, input_map=None, input_options=None, use_hw=False, quality="default"):
+    def start(self, host, name, profile, resolution='720p', fps=60, show_fps=False, fullscreen=False, input_map=None, input_options=None, use_hw=False, quality="default", audio_device=None, use_opengl=False):
+        self.center_text.show()
         self.input_options = input_options
-        self.frame_mutex = QtCore.QMutex()
-        self.video_output.hide()
         self.mapping = ControlsWidget.DEFAULT_MAPPING if input_map is None else input_map
         self.fps = fps
         self.fullscreen = fullscreen
-        self.ms_refresh = 1000.0/self.fps
+        self.use_opengl = use_opengl
+        self.ms_refresh = 1000.0/self.fps * 0.4
         self.setWindowTitle(f"Session {name} @ {host}")
         self.rp_worker.setup(self, host, profile, resolution, fps, use_hw, quality)
+        self.resize(self.rp_worker.session.resolution['width'], self.rp_worker.session.resolution['height'])
+        output = YUVGLWidget if self.use_opengl else VideoWidget
+        self.video_output = output(self.rp_worker.session.resolution['width'], self.rp_worker.session.resolution['height'])
+        self.video_output.hide()
+        self.layout.addWidget(self.video_output)
+        self.joystick.setParent(self.video_output)
+        self.fps_label.setParent(self.video_output)
         self.show_fps = show_fps
+        self.audio_device = audio_device
 
         if self.show_fps:
             self.init_fps()
             self.fps_label.show()
         else:
             self.fps_label.hide()
-        self.av_thread.start()
         self.started.connect(self.main_window.session_start)
         self.started.emit()
-
-    def show_video(self):
-        self.resize(self.rp_worker.session.resolution['width'], self.rp_worker.session.resolution['height'])
         if self.fullscreen:
             self.showFullScreen()
         else:
             self.show()
+
+    def show_video(self):
+        self.center_text.hide()
         self.video_output.show()
         joysticks = self.input_options.get("joysticks")
         if joysticks:
             self.joystick.hide_sticks()
             self.joystick.show_sticks(joysticks['left'], joysticks['right'])
             self.joystick.default_pos()
+        self.setFixedSize(self.width(), self.height())
 
-# Waiting on pyside6.2
-#    def init_audio(self):
-#        config = self._a_stream._audio_config
-#        format = QtMultimedia.QAudioFormat()
-#        format.setChannels(config['channels'])
-#        format.setFrequency(config['rate'])
-#        format.setSampleSize(config['bits'])
-#        format.setCodec("audio/pcm")
-#        format.setByteOrder(QtMultimedia.QAudioFormat.LittleEndian)
-#        format.setSampleType(QtMultimedia.QAudioFormat.SignedInt)
-#        output = QtMultimedia.QAudioOutput(format)
-#        self.audio_output = output.start()
-
-    def new_frame(self):
-        try:
-            pixmap = self.av_worker.pixmaps.popleft()
-        except IndexError:
+    def init_audio(self):
+        if not self.rp_worker.session:
             return
-        if self.fullscreen:
-            pixmap = pixmap.scaled(self.video_output.size(), aspectMode=Qt.KeepAspectRatio, mode=self._video_transform_mode)
-        self.video_output.setPixmap(pixmap)
-        if self.show_fps:
-            self.set_fps()
+        config = self.rp_worker.session.av_receiver.audio_config
+        audio_format = QtMultimedia.QAudioFormat()
+        audio_format.setChannelCount(config['channels'])
+        audio_format.setSampleRate(config['rate'])
+        audio_format.setSampleFormat(QtMultimedia.QAudioFormat.Int16)
 
-    def av_slow(self):
-        self._video_transform_mode = Qt.FastTransformation
+        self.audio_output = QtMultimedia.QAudioSink(self.audio_device, format=audio_format)
+        self.audio_output.setBufferSize(3840)
+        self.audio_buffer = self.audio_output.start()
+
+    def next_audio_frame(self, data):
+        if self.audio_output is None:
+            self.init_audio()
+        if self.audio_buffer:
+            self.audio_buffer.write(data)
 
     def init_fps(self):
         self.fps_label.move(20, 20)
+        self.fps_label.resize(80, 20)
         self.fps_label.setStyleSheet("background-color:#33333333;color:white;padding-left:5px;")
         self.fps_sample = 0
         self.last_time = time.time()
@@ -396,11 +421,6 @@ class StreamWindow(QtWidgets.QWidget):
             self.fps_label.setText(f"FPS: {int(self.fps/delta)}")
             self.fps_sample = 0
 
-    def handle_audio(self, data):
-        if self.audio_output is None:
-            self.init_audio()
-        self.audio_output.write()
-
     def mousePressEvent(self, event):
         button = event.button().name.decode()
         self._handle_press(button)
@@ -469,11 +489,12 @@ class StreamWindow(QtWidgets.QWidget):
 
     def cleanup(self):
         print("Cleaning up window")
-        self.timer.stop()
-        self.av_thread.quit()
-        event_emitter.remove_all_listeners("frame")
+        event_emitter.remove_all_listeners()
+        if self.video_output:
+            self.video_output.deleteLater()
+            self.video_output = None
+        if self.audio_output:
+            self.audio_output.stop()
+            self.audio_output = None
+            self.audio_buffer = None
         self.main_window.session_stop()
-
-    def start_timer(self):
-        print("AV Processor Started")
-        self.timer.start(self.ms_refresh)
