@@ -11,18 +11,17 @@ from struct import pack_into
 
 import requests
 from Cryptodome.Random import get_random_bytes
-from pyps4_2ndscreen.ddp import get_status
 
-from .av import AVFileReceiver, AVHandler
-from .const import (DDP_PORT, DDP_VERSION, FPS, OS_TYPE, RP_CRYPT_SIZE,
-                    RP_PORT, RP_VERSION, TYPE_PS4, TYPE_PS5, USER_AGENT,
-                    Resolution)
+from .av import AVHandler
+from .const import (FPS, OS_TYPE, RP_CRYPT_SIZE, RP_PORT, RP_VERSION, TYPE_PS4,
+                    TYPE_PS5, USER_AGENT, Resolution)
 from .crypt import SessionCipher
+from .ddp import get_status, wakeup
 from .errors import RemotePlayError, RPErrorHandler
 from .feedback import Controller
 from .keys import SESSION_KEY_0, SESSION_KEY_1
 from .stream import RPStream
-from .util import listener, log_bytes
+from .util import format_regist_key, listener, log_bytes
 
 _LOGGER = logging.getLogger(__name__)
 
@@ -102,30 +101,6 @@ def _gen_did() -> bytes:
     return did
 
 
-def get_wakeup_packet(regist_key: str) -> bytes:
-    regist_key = int.from_bytes(bytes.fromhex(bytes.fromhex(regist_key).decode()), "big")
-
-    data = (
-        "WAKEUP * HTTP/1.1\n"
-        "client-type:vr\n"
-        "auth-type:R\n"
-        "model:w\n"
-        "app-type:r\n"
-        f"user-credential:{regist_key}\n"
-        f"device-discovery-protocol-version:{DDP_VERSION}\n"
-    )
-    return data.encode()
-
-
-def send_wakeup(host: str, regist_key: str):
-    """Send Wakeup Packet."""
-    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
-    sock.settimeout(0)
-    sock.sendto(get_wakeup_packet(regist_key), (host, DDP_PORT))
-    _LOGGER.info("Sent Wakeup to host")
-    sock.close()
-
-
 class Session():
     """Controller for RP Session."""
     STATE_INIT = "init"
@@ -185,6 +160,7 @@ class Session():
 
         self._ready_event = None
         self._stop_event = None
+        self.stream_ready = None
         self.receiver_started = None
 
     def _init_profile_kwargs(self, device: dict) -> bool:
@@ -219,7 +195,7 @@ class Session():
 
     def _get_status(self) -> dict:
         """Return dict of device status."""
-        return get_status(self._host)
+        return get_status(self._host, host_type=self.type)
 
     def _check_host(self) -> tuple:
         """Return True, True if host is available."""
@@ -380,13 +356,15 @@ class Session():
 
     def wakeup(self):
         """Wakeup Host."""
-        send_wakeup(self._host, self._regist_key)
+        regist_key = format_regist_key(self._regist_key)
+        wakeup(self.host, regist_key, host_type=self.type)
 
     def start(self, wakeup=True, autostart=True) -> bool:
         """Start Session/RP Session."""
         self._ready_event = threading.Event()
         self._stop_event = threading.Event()
         self.receiver_started = threading.Event()
+        self.stream_ready = threading.Event()
         status = self._check_host()
         if not status[0]:
             self.error = f"Host @ {self._host} is not reachable."
@@ -454,7 +432,8 @@ class Session():
             _LOGGER.debug("Session already stopping")
             return
         _LOGGER.info("Session Received Stop Signal")
-        self._stop_event.set()
+        if self._stop_event:
+            self._stop_event.set()
 
     def init_controller(self):
         self.controller.start()
@@ -506,10 +485,10 @@ class Session():
     @property
     def video_format(self):
         formats = {
-            "ps4": "h264",
-            "ps5": "h265",
+            TYPE_PS4: "h264",
+            TYPE_PS5: "h265",
         }
-        return formats.get(self.type.lower())
+        return formats.get(self.type)
     
 
 class SessionAsync(Session):
@@ -550,6 +529,7 @@ class SessionAsync(Session):
         self._ready_event = asyncio.Event()
         self._stop_event = asyncio.Event()
         self.receiver_started = asyncio.Event()
+        self.stream_ready = asyncio.Event()
 
         _LOGGER.debug("Running Async")
         status = await self.run_io(self._check_host)
