@@ -57,6 +57,12 @@ class RPStream():
         def close(self):
             self.transport.close()
 
+        @property
+        def socket(self):
+            if self.transport is None:
+                return None
+            return self.transport.get_extra_info("socket")
+
     def __init__(self, session, stop_event, rtt=None, mtu=None, is_test=False, cb_stop=None):
         self._host = session.host
         self._port = STREAM_PORT if not is_test else TEST_STREAM_PORT
@@ -109,6 +115,7 @@ class RPStream():
     def ready(self):
         _LOGGER.info("Stream Ready")
         self._state = RPStream.STATE_READY
+        self._session.stream_ready.set()
 
     def _advance_sequence(self):
         """Advance SCTP sequence number."""
